# coding: utf-8

from __future__ import absolute_import
from __future__ import unicode_literals

from django.contrib import auth
from django.core.exceptions import ImproperlyConfigured
from m3_auth.backends.base import BaseBackend
from m3_auth.exceptions import AuthError
from m3_django_compat import get_request_params
from m3_django_compat import get_user_model


try:
    from django_auth_ldap.backend import LDAPSettings
    from django_auth_ldap.backend import LDAPBackend
except ImportError:
    raise ImproperlyConfigured(
        'LdapBackend depends on django-auth-ldap>=1.7<1.8'
        'You need to check your environment.'
    )


class LdapBackend(BaseBackend):
    """Бэкенд для аутентификации по протоколу Ldap."""

    def __init__(self, name, config):
        super(LdapBackend, self).__init__(name)
        self.config = config
        settings = LDAPSettings(
            defaults=dict(
                SERVER_URI=config.server_uri,
                BIND_DN=config.bind_dn,
                BIND_PASSWORD=config.bind_password,
                USER_DN_TEMPLATE=config.user_dn_template,
                USER_ATTR_MAP=config.user_attr_map,
            )
        )
        self._ldap_backend = LDAPBackend()
        self._ldap_backend.settings = settings

    def process_login(self, request):
        """Обработка запроса аутенфикации.

        При помощи класса LDAPBackend осщуествляется поиск пользователя в AD.
        Если пользователь найден то происходит его аутенфикация при помощи
        модуля django.auth.
        """

        user_model = get_user_model()
        username_field = getattr(user_model, 'USERNAME_FIELD', 'username')

        params = get_request_params(request)
        username = params.get(username_field)
        password = params.get('password')

        if not all((username, password)):
            model_meta = user_model._meta
            raise AuthError(
                'Отсутствуют обязательные параметры: "{}" и/или "{}".'.format(
                    model_meta.get_field(username_field).verbose_name,
                    model_meta.get_field('password').verbose_name
                ))

        user = self._ldap_backend.authenticate(
            request, username=username, password=password
        )
        if not user:
            raise AuthError('Пользователь не найден в Active Directory.')
        else:
            auth.login(request, user)

        return user

    def get_login_url(self, request):
        """Возращает True т.к. нет необходимости в перенаправлении."""
        return True

    def get_logout_url(self, request):
        """Возращает True т.к. нет необходимости в перенаправлении."""
        return True

    def process_logout(self, request):
        """Возвращает True, т.к. logout осуществляется средствами django."""
        return True

    @classmethod
    def from_config(cls, config):
        """Метод создает и возвращает экземпляр класса бэкенда на основе
        параметров config.

        В конфиге должны быть два обязательных параметра:
        1. name - наименование бэкенда.
        2. config - Конфиги используемые самим бэкендом.

        :param config: Конфигурация бэкенда.
        :type config: auth.config.Config
        :return: Экземпляр класса cls.
        :rtype: m3_auth.backends.ldap.LdapBackend
        """

        return cls(config.name, config.config)
