# coding: utf-8

from __future__ import absolute_import
from __future__ import unicode_literals

from django.contrib.auth.middleware import AuthenticationMiddleware
from django.contrib.sessions.middleware import SessionMiddleware
from django.contrib.auth.models import User
from django.db.models.query import QuerySet
from django.test.client import RequestFactory
from django.test import TestCase

from m3_auth.backends.base.remote import RemoteBackend
from m3_auth.backends.base.remote.protocol import BaseProtocol
from m3_auth.backends.base.remote.user import BaseRemoteUser
from m3_auth.config import Config
from m3_auth.exceptions import AuthError
from m3_auth.exceptions import BadRequest
from m3_auth.exceptions import MoreUsersFound
from m3_auth.exceptions import UserNotFound
from m3_auth.fields import DateField
from m3_auth.fields import SimpleField
from m3_auth.proxy import SessionProxy
from m3_auth.utils import import_object
from m3_auth.utils import is_authenticated


class TestProtocol(BaseProtocol):
    """Тестовый протокол, для проверки RemoteBackend."""

    def get_login_url(self, request):
        return self.config.login_url

    def process_login(self, request):
        return self.user_cls(self.config.user_data)

    def get_logout_url(self, request):
        return self.config.logout_url

    def process_logout(self, request):
        return True

    @classmethod
    def from_config(cls, config):
        user_cls = import_object(config.user_cls)
        return cls(user_cls, config.config)


def fullname_filter(request, queryset, remote_user):
    """Фильтрация по ФИО."""
    return queryset.filter(
        person__surname=remote_user.surname,
        person__first_name=remote_user.first_name,
        person__patronymic=remote_user.patronymic
    )


def date_of_birth_filter(request, queryset, remote_user):
    """Фильтрация по дате рождения."""
    return queryset.filter(person__date_of_birth=remote_user.date_of_birth)


class TestRemoteUser(BaseRemoteUser):

    """Тестовый класс пользователя."""

    surname = SimpleField('surname')
    first_name = SimpleField('first_name')
    patronymic = SimpleField('patronymic')
    date_of_birth = DateField('date_of_birth', '%d.%m.%Y')


class TestRemoteBackend(TestCase):

    """Проверка базового бэкенда RemoteBackend."""

    def setUp(self):
        super(TestRemoteBackend, self).setUp()
        self.factory = RequestFactory()
        self.config = Config({
            'login_url': 'http://localhost:8000/process_login',
            'logout_url': 'http://localhost:8000/process_logout'
        })
        self.user_data = {
            'user_id': 123,
            'surname': 'Супер',
            'first_name': 'Пользователь',
            'patronymic': 'Системы',
            'date_of_birth': '01.01.1990'
        }

    def create_backend(self, filters=None):
        """Создает бэкенд, с разными фильтрами."""
        protocol = TestProtocol(TestRemoteUser, self.config)
        return RemoteBackend('remote', protocol, filters)

    @staticmethod
    def apply_middleware(request):
        SessionMiddleware().process_request(request)
        AuthenticationMiddleware().process_request(request)

    def process_login(self, user_data, filters, request=None):
        """Метод для проверки входа."""
        request = request or self.factory.post('/')
        backend = self.create_backend(filters)

        self.apply_middleware(request)
        self.config.user_data = user_data

        result = backend.process_login(request)
        self.config.user_data = None
        return request, result

    def login_with_selected_user(
        self, request, token=None, user_ids=None, params=None
    ):
        """Вход под выбранным пользователем."""
        backend = self.create_backend([date_of_birth_filter])
        session_proxy = SessionProxy(request)
        session_proxy.token = token
        session_proxy.user_ids_to_select = user_ids

        new_request = self.factory.post('/', params)
        new_request.session = request.session
        new_request.user = request.user
        return backend.login_with_selected_user(new_request), new_request

    def test_get_login_url(self):
        """Проверка получения URL адреса на вход через протокол."""
        request = self.factory.post('/')
        backend = self.create_backend()
        login_url = backend.get_login_url(request)
        self.assertEqual(login_url, self.config.login_url)

    def test_process_login_with_one_user(self):
        """Проверка если найден 1 пользователь."""

        # fullname_filter должен вернуть queryset только с одним пользователем.
        request, user = self.process_login(self.user_data, [fullname_filter])
        session_proxy = SessionProxy(request)

        self.assertIsInstance(user, User)
        self.assertTrue(is_authenticated(request))
        self.assertEqual(
            session_proxy.remote_user_id, self.user_data['user_id'])
        self.assertEqual(session_proxy.backend, 'remote')

    def test_process_login_without_existing_user(self):
        """Проверка входа под не существующим пользователем."""
        self.assertRaises(UserNotFound, self.process_login, {
            'user_id': 1, 'surname': 'SomeUndefinedSurname'
        }, [fullname_filter])

    def test_process_login_with_several_users(self):
        """Проверка входа когда найдено несколько пользователей."""
        request = self.factory.post('/')

        # date_of_birth_filter должен найти несколько пользователей.
        # это Супер пользователь системы и Не активный пользователь, с
        # одинаковой датой рождения
        self.assertRaises(
            MoreUsersFound, self.process_login, self.user_data,
            [date_of_birth_filter], request
        )

        session_proxy = SessionProxy(request)

        self.assertTrue(session_proxy.remote_user_id)
        self.assertTrue(session_proxy.token)
        self.assertTrue(session_proxy.user_ids_to_select)

    def test_login_with_several_users(self):
        """Проверка входа под выбранным пользователем."""
        request = self.factory.post('/')

        # Перехватываем исключение
        self.assertRaises(
            MoreUsersFound, self.process_login, self.user_data,
            [date_of_birth_filter], request
        )

        admin = User.objects.get(username='admin')
        session_proxy = SessionProxy(request)

        # Так как эти параметры затираются, то надо каждый раз записывать их
        # в session_proxy
        token = session_proxy.token
        user_ids = session_proxy.user_ids_to_select

        # Проверка если токен потрачен
        session_proxy.token = None
        self.assertRaises(BadRequest, self.login_with_selected_user, request)
        session_proxy.token = token

        # Проверка если пользователь уже вошел в систему
        anonymous_user = request.user
        request.user = admin

        self.assertRaises(
            AuthError, self.login_with_selected_user, request, token)
        request.user = anonymous_user

        # Проверка если не переданы параметры token, user_id
        self.assertRaises(BadRequest, self.login_with_selected_user, request)

        # Проверка затерлись ли параметры
        self.assertFalse(session_proxy.token)
        self.assertFalse(session_proxy.user_ids_to_select)

        # Проверка если user_id is not digit.
        self.assertRaises(
            BadRequest, self.login_with_selected_user, request, token,
            user_ids, {'token': token, 'user_id': 'f'})

        # Проверка если токен не верный.
        self.assertRaises(
            BadRequest, self.login_with_selected_user, request, token,
            user_ids, {'token': 'some_token', 'user_id': '1'})

        # Проверка, если такой пользователь не найден
        self.assertRaises(
            BadRequest, self.login_with_selected_user, request, token,
            user_ids, {'token': token, 'user_id': '123456'})

        # Проверка если id не входит в список предоставленного выбора
        self.assertRaises(
            BadRequest, self.login_with_selected_user, request, token, [1],
            {'token': token, 'user_id': '2'})

        # Проверка если пользователь не активен
        self.assertRaises(
            AuthError, self.login_with_selected_user, request, token,
            user_ids, {'token': token, 'user_id': '2'})

        # Проверка если все ОК
        user, good_request = self.login_with_selected_user(
            request, token, user_ids, {'token': token, 'user_id': '1'})

        self.assertTrue(user)
        self.assertIsInstance(user, User)
        self.assertTrue(is_authenticated(good_request))

    def test_apply_filters(self):
        """Проверка метода apply_filters."""
        request = self.factory.post('/')
        queryset = User.objects.all()
        user = TestRemoteUser(self.user_data)

        # Если фильтры не указаны
        backend = self.create_backend()
        result = backend.apply_filters(request, queryset, user)

        self.assertIsInstance(result, QuerySet)
        self.assertFalse(result.count())

        # Если указан 1 фильтр.
        backend = self.create_backend([date_of_birth_filter])
        result = backend.apply_filters(request, queryset, user)

        self.assertIsInstance(result, QuerySet)
        self.assertEqual(result.count(), 2)

        # Если указано 2 фильтра
        backend = self.create_backend([date_of_birth_filter, fullname_filter])
        result = backend.apply_filters(request, queryset, user)

        self.assertIsInstance(result, QuerySet)
        self.assertEqual(result.count(), 1)

    def test_login_method(self):
        """Проверка метода _login."""
        request = self.factory.post('/')
        self.apply_middleware(request)

        session_proxy = SessionProxy(request)
        user = User.objects.get(username='admin')
        remote_user_id = 'remote_user_id'

        backend = self.create_backend()
        backend._login(request, user, remote_user_id)

        self.assertTrue(is_authenticated(request))
        self.assertEqual(request.user, user)

        self.assertEqual(session_proxy.backend, backend.name)
        self.assertEqual(session_proxy.remote_user_id, remote_user_id)

    def test_logout_method(self):
        """Проверка метода _logout."""
        request, user = self.process_login(self.user_data, [fullname_filter])
        session_proxy = SessionProxy(request)

        self.assertTrue(user)
        self.assertTrue(is_authenticated(request))

        # Выходим из системы
        backend = self.create_backend()
        backend._logout(request)

        self.assertFalse(is_authenticated(request))
        self.assertFalse(session_proxy.remote_user_id)
        self.assertFalse(session_proxy.backend)

    def test_process_logout(self):
        """Проверка выхода из системы."""

        request, user = self.process_login(self.user_data, [fullname_filter])
        self.assertTrue(is_authenticated(request))

        backend = self.create_backend([])
        backend.process_logout(request)

        session_proxy = SessionProxy(request)
        self.assertFalse(is_authenticated(request))
        self.assertFalse(session_proxy.backend)
        self.assertFalse(session_proxy.remote_user_id)

    def test_get_logout_url(self):
        """Проверка получения URL адреса на выход через протокол."""
        request = self.factory.post('/')
        backend = self.create_backend()
        logout_url = backend.get_logout_url(request)
        self.assertEqual(logout_url, self.config.logout_url)

    def test_from_config(self):
        """Проверка метода from_config."""
        config = Config({
            'name': 'remote',
            'user_filters': [
                '.'.join((__name__, fullname_filter.__name__)),
                '.'.join((__name__, date_of_birth_filter.__name__))
            ],
            'protocol': {
                'protocol_cls': '.'.join((__name__, TestProtocol.__name__)),
                'user_cls': '.'.join((__name__, TestRemoteUser.__name__)),
                'config': {
                    'login_url': 'login_url',
                    'logout_url': 'logout_url'
                }
            }
        })

        backend = RemoteBackend.from_config(config)

        self.assertEqual(backend.name, 'remote')
        self.assertEqual(backend.filters, [
            fullname_filter, date_of_birth_filter
        ])
        self.assertIsInstance(backend.protocol, TestProtocol)
        self.assertIs(backend.protocol.user_cls, TestRemoteUser)
        self.assertIs(backend.protocol.config, config.protocol.config)

        # Проверка без указания фильтров
        config.user_filters = []
        backend = RemoteBackend.from_config(config)
        self.assertEqual(backend.filters, [])
